/*
 *  Copyright (C) 2007 Free Software Foundation, Inc.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define CODE_ADDR	0x7C00
#define CODE_LENG	(code_end - start)	// 1 sector
#define DATA_ADDR	0x8200

#define BLCK_LENG	0x4000

        .file   "lnxhdr.S"

        .code16

        .globl  start, _start

	. = . + 0x1F1
setup_sects:
	.byte	(CODE_LENG >> 9)
root_flags:
	.word	0
syssize:
	.word	0
swap_dev:
	.word	0
ram_size:
	.word	0
vid_mode:
	.word	0
root_dev:
	.word	0
boot_flag:
	.word	0xAA55

start:
_start:

	jmp code_start

	.ascii	"HdrS"			// Header signature
	.word	0x0203			// Header version number

realmode_swtch:
	.word	0, 0			// default_switch, SETUPSEG
start_sys_seg:
	.word	0x1000			// obsolete
version_ptr:
	.word	version - start		// version string ptr
type_of_loader:
	.byte	0			// Filled in by boot loader
loadflags:
	.byte	1			// Please load high
setup_move_size:
	.word	0			// Unused
code32_start:
	.long	0x100000		// 32-bit start address
ramdisk_image:
	.long	0			// Loaded ramdisk image address
ramdisk_size:
	.long	0			// Size of loaded ramdisk
bootsect_kludge:
	.word	0, 0
heap_end_ptr:
	.word	0
pad1:
	.word	0
cmd_line_ptr:
	.long	0			// Command line
ramdisk_max:
	.long	0xffffffff		// Highest allowed ramdisk address

gdt:
	.long	0, 0, 0, 0		// Must be zero
	.word	0xffff			// 64 K segment size
gdt_src1:
	.byte	0, 0 ,0			// Low 24 bits of source addy
	.byte	0x93			// Access rights
	.byte	0			// Extended access rights
gdt_src2:
	.byte	0			// High 8 bits of source addy
	.word	0xffff			// 64 K segment size
gdt_dst1:
	.byte	0, 0, 0			// Low 24 bits of target addy
	.byte	0x93			// Access rights
	.byte	0			// Extended access rights
gdt_dst2:
	.byte	0			// High 8 bits of source addy
	.long	0, 0, 0, 0		// More space for the BIOS

code_start:

	// Setup stack

	xorw	%ax, %ax
	movw	%ax, %ss
	movw	$(CODE_ADDR), %sp

	pushl	%esi
	pushl	%edi
	pushl	%edx

	// Move itself to 0:CODE_ADDR

	cld
	movw	%cs, %ax
	movw	%ax, %ds
	movw	$(CODE_ADDR >> 4), %ax
	movw	%ax, %es
	xorw	%si, %si
	movw	%si, %di
	movw	$(CODE_LENG >> 2), %cx

	rep
	movsl

	ljmp	$0, $(jump_start - start + CODE_ADDR)

jump_start:

	xorw	%ax, %ax
	movw	%ax, %ds
	movw	%ax, %es

	movl	$0x200, %ecx
	movl	(code32_start - start + CODE_ADDR), %esi
	addl	%ecx, %esi
	movl	$DATA_ADDR, %edi

	call	move_memory

	popl	%edx
	pushl	%edx

	xorl	%eax, %eax
	movb	%dh, %al
	movl	%eax, (0x14 + DATA_ADDR)	// GRUB_KERNEL_MACHINE_INSTALL_DOS_PART
	movl	(0x10 + DATA_ADDR), %ecx	// GRUB_KERNEL_MACHINE_COMPRESSED_SIZE
	addl	$(0x4A0 - 0x200), %ecx		// GRUB_KERNEL_MACHINE_RAW_SIZE
	incl	%ecx
	andb	$0xFE, %cl

	call	move_memory

	popl	%edx
	popl	%esi
	popl	%edi

	ljmp	$0, $(DATA_ADDR)

// Parameters:
//   esi: source address
//   edi: target address
//   ecx: number of bytes
move_memory:
1:
	pushl	%esi
	pushl	%edi
	pushl	%ecx
	cmpl	$BLCK_LENG, %ecx
	jbe	2f
	movl	$BLCK_LENG, %ecx
2:
	pushl	%ecx

	movl	%esi, %eax
	movw	%si, (gdt_src1 - start + CODE_ADDR)
	shrl	$16, %eax
	movb	%al, (gdt_src1 + 2 - start + CODE_ADDR)
	movb	%ah, (gdt_src2 - start + CODE_ADDR)

	movl	%edi, %eax
	movw	%di, (gdt_dst1 - start + CODE_ADDR)
	shrl	$16, %eax
	movb	%al, (gdt_dst1 + 2 - start + CODE_ADDR)
	movb	%ah, (gdt_dst2 - start + CODE_ADDR)

	movw	$(gdt - start + CODE_ADDR), %si
	movb	$0x87, %ah
	shrw	$1, %cx

	int	$0x15

	popl	%eax
	popl	%ecx
	popl	%edi
	popl	%esi

	jnc	2f
	movw	$(err_int15_msg - start + CODE_ADDR), %si
	jmp	fail
2:

	leal	(%esi, %eax), %esi
	leal	(%edi, %eax), %edi
	subl	%eax, %ecx
	jnz	1b

	ret

// Parameters:
//   si: error message
fail:
	movb	$0x0e, %ah
	xorw	%bx, %bx
1:
	lodsb	(%si), %al
	int	$0x10
	cmpb	$0, %al
	jne	1b
1:	jmp	1b

version:
	.ascii	"lnxhdr\0"

err_int15_msg:
	.ascii	"move memory fails\0"


	. = (.  & (~0x1FF)) + 0x1FF

	.byte	0

code_end:
